# -*- coding:utf8 -*-

import json
import numpy as np


def softmax(x):
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum()


class Wordlist(object):
    def __init__(self, filename):
        lines = map(lambda x: x.split(), open(filename).readlines())
        self._size = len(lines)

        voc = [(item[1], item[0][0]) for item in zip(lines, xrange(self._size))]
        self._voc = dict(voc)

    def get_word(self, word_id):
        if word_id in self._voc:
            return self._voc[word_id]
        else:
            return None

    @property
    def size(self):
        return self._size


def genAttention(filename, savingpath, voc):
    data = json.load(open(filename, "r"))
    docs = data["docs"]
    sent_length = data["length"]
    batch_size = len(data["sentence_mask"][0])
    sentiment_lexicon = dict(json.load(open("./data/yelp14/Attention_Lexicon.json", "r")))
    word_degree = []
    sent_degree = []
    word_count = 0.0
    senti_count = 0.0
    for i in range(len(docs)):
        batch_sent_attention = []
        batch_doc_attention = []
        doc_tmp = []
        for j in range(len(docs[i])):
            sent_attention = []
            sentence = []
            for word_id in docs[i][j]:
                word = voc.get_word(word_id - 1)
                if word is None:
                    sent_attention.append(float("-Inf"))
                elif word.lower() in sentiment_lexicon:
                    sent_attention.append(2.0 * sentiment_lexicon[word.lower()]) # the lambda, 2.0 or 3.0
                    sentence.append(word.lower())
                    word_count += 1
                    senti_count += 1
                else:
                    sent_attention.append(0.0)
                    sentence.append(word.lower())
                    word_count += 1
            # print(sent_senti_mask)
            if sent_attention[0] == float("-Inf"):
                sent_attention_softmax = np.zeros(len(sent_attention), dtype=np.float32).tolist()
            else:
                sent_attention_softmax = softmax(np.array(sent_attention)).tolist()
            # print(sentence)
            # print(sent_attention_softmax)
            batch_sent_attention.append(sent_attention_softmax)

            assert sent_length[i][j] > 0
            doc_tmp.append(3.0 * np.sum(sent_attention) / sent_length[i][j]) # The lambda, 2.0 or 3.0
            if len(doc_tmp) == len(docs[i]) / batch_size:
                if doc_tmp[0] == float("-Inf"):
                    doc_tmp = np.zeros(len(doc_tmp), dtype=np.float32).tolist()
                else:
                    doc_tmp = softmax(np.array(doc_tmp)).tolist()
                batch_doc_attention.append(doc_tmp)
                doc_tmp = []
        word_degree.append(batch_sent_attention)
        sent_degree.append(batch_doc_attention)

    print(senti_count / word_count)
    dict_t = {"word_degree": word_degree,
              "sent_degree": sent_degree}

    json.dump(dict_t, open(savingpath, "wb"))


def attentionProcessing():
    voc = Wordlist("/media/rowitzou/WORK/NLP/sentiment-classification/data/yelp14/wordlist.txt")
    genAttention("./data/yelp14/train_32.json", "./data/yelp14/supervised_attention_32.json", voc)


attentionProcessing()
