# -*- coding:utf8 -*-

from LSTMModel import LSTMModel
from Config import Config
import tensorflow as tf
import numpy as np
import json
import copy


class Wordlist(object):
    def __init__(self, filename):
        lines = list(map(lambda x: x.split(), open(filename).readlines()))
        self._size = len(lines)

        voc = [(item[1], item[0][0]) for item in zip(lines, range(self._size))]
        self._voc = dict(voc)

    def get_word(self, word_id):
        if word_id in self._voc:
            return [item for item in self._voc[word_id]]
        else:
            return []

    @property
    def size(self):
        return self._size


class Charlist(object):

    def __init__(self, filename):
        line = open(filename).readline()
        self._size = len(line)
        line = [item for item in line]

        voc = [(item[0], item[1]) for item in zip(line, range(self._size))]
        self._voc = dict(voc)

    def get_id(self, char):
        if char in self._voc:
            return self._voc[char]
        else:
            return -1

    @property
    def size(self):
        return self._size


def run_epoch(session, model, data, eval_op, vali=False, attention_data=None):

    accuracy = 0.0
    ave_loss = 0.0
    ave_c_loss = 0.0
    ave_w_a_loss = 0.0
    ave_s_a_loss = 0.0
    ave_mse = 0.0

    # char_voc = Charlist("./data/yelp13/charlist.txt")
    # voc = Wordlist("./data/yelp13/wordlist.txt")
    shuffle_id = np.arange(len(data["docs"]))
    np.random.shuffle(shuffle_id)

    for i in range(len(data["docs"])):

        if not vali:
            fetches = [eval_op,
                       model.accuracy,
                       model.loss,
                       model.mse,
                       model.classification_loss,
                       model.word_attention_loss,
                       model.sent_attention_loss]
        else:
            fetches = [eval_op, model.accuracy, model.mse]

        feed_dict = {}
        feed_dict[model.input] = data["docs"][shuffle_id[i]]
        feed_dict[model.label] = data["labels_one_hot"][shuffle_id[i]]
        feed_dict[model.sent_mask] = data["sentence_mask"][shuffle_id[i]]
        feed_dict[model.word_mask] = data["word_mask"][shuffle_id[i]]
        feed_dict[model.sent_sequence_length] = data["sentence_num"][shuffle_id[i]]
        feed_dict[model.word_sequence_length] = data["length"][shuffle_id[i]]

        if not (attention_data is None):
            feed_dict[model.supervised_attention_word] = attention_data["word_degree"][shuffle_id[i]]
            feed_dict[model.supervised_attention_sent] = attention_data["sent_degree"][shuffle_id[i]]

        word_softmax_mask = np.array(data["word_mask"][shuffle_id[i]], dtype=np.float32)
        word_softmax_mask[word_softmax_mask < 0.9] = float('-Inf')
        word_softmax_mask[word_softmax_mask > 0.9] = float(0.0)
        feed_dict[model.word_soft_max] = word_softmax_mask

        sent_softmax_mask = np.array(data["sentence_mask"][shuffle_id[i]], dtype=np.float32)
        sent_softmax_mask[sent_softmax_mask < 0.9] = float('-Inf')
        sent_softmax_mask[sent_softmax_mask > 0.9] = float(0.0)
        feed_dict[model.sent_soft_max] = sent_softmax_mask

        """
        docs_char = copy.deepcopy(data["docs"][shuffle_id[i]])
        docs_char = list(map(lambda x: list(map(lambda y: list(map(lambda ch: char_voc.get_id(ch) + 1, voc.get_word(y - 1))), x)),
                        docs_char))
        docs_char = list(map(lambda x: list(map(lambda y: [y, y[:config.max_word_length - 1]][len(y) > config.max_word_length], x)), docs_char))
        docs_char = list(map(lambda x: list(map(lambda y: y + [0] * (config.max_word_length - len(y)), x)), docs_char))
        feed_dict[model.input_char] = docs_char
        """
        if not vali:
            _, acc, loss, mse, c_loss, w_a_loss, s_a_loss = session.run(fetches, feed_dict=feed_dict)
            ave_loss += loss
            ave_c_loss += c_loss
            ave_w_a_loss += w_a_loss * config.word_attention_loss_lambda
            ave_s_a_loss += s_a_loss * config.sent_attention_loss_lambda
        else:
            _, acc, mse = session.run(fetches, feed_dict=feed_dict)

        accuracy += acc
        ave_mse += mse

        if i > 0 and i % 20 == 0:
            if vali:
                print("%d Accuracy: %lf, mse: %lf" %
                      (i // 20, (accuracy / (i + 1)), np.sqrt(ave_mse / (i + 1))))
            else:
                # print(len(cnn_output[0][0]))
                print("%d Train acc: %lf, mse: %lf, loss: %lf, c_loss: %lf, w_a_loss: %lf, s_a_loss: %lf" %
                      (i // 20, (accuracy / (i + 1)), np.sqrt(ave_mse / (i + 1)), ave_loss / (i + 1),
                       ave_c_loss / (i + 1), ave_w_a_loss / (i + 1), ave_s_a_loss / (i + 1)))

    if vali:
        print("Final accuracy: %lf." % (accuracy / len(data["docs"])))
        print("Final RMSE: %lf." % np.sqrt(ave_mse / len(data["docs"])))
        return accuracy / len(data["docs"]), np.sqrt(ave_mse / len(data["docs"]))


def train(train_data, val_data, test_data, config, path, attention_data=None, max_epoch=20):

    gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=0.8)
    with tf.Graph().as_default(), tf.Session(config=tf.ConfigProto(gpu_options=gpu_options)) as session:

        initializer = tf.random_uniform_initializer(-0.1, 0.1)

        with tf.variable_scope("model", reuse=None, initializer=initializer):
            model_train = LSTMModel(config, is_training=True)

        with tf.variable_scope("model", reuse=True, initializer=initializer):
            model_val = LSTMModel(config, is_training=False)

        tf.global_variables_initializer().run()
        
        max_accuracy = 0.0
        record_epoch = 0.0
        best_rmse = 0.0
        for i in range(max_epoch):

            print("Training Epoch: %d" % (i + 1))
            run_epoch(session, model_train, train_data, model_train.train_op, attention_data=attention_data)

            print("Validation:")
            acc, rmse = run_epoch(session, model_val, val_data, tf.no_op(), vali=True)

            if acc > max_accuracy:
                max_accuracy = acc
                best_rmse = rmse
                record_epoch = i + 1
                model_val.input_hidden_layer_saver.save(session, "./model/" + path + "/input_hidden_layer.ckpt")
                model_val.sent_lstm_saver.save(session, "./model/" + path + "/sent_lstm.ckpt")
                model_val.sent_attention_saver.save(session, "./model/" + path + "/sent_attention.ckpt")
                model_val.doc_lstm_saver.save(session, "./model/" + path + "/doc_lstm.ckpt")
                model_val.doc_attention_saver.save(session, "./model/" + path + "/doc_attention.ckpt")
                model_val.output_hidden_layer_saver.save(session, "./model/" + path + "/output_hidden_layer.ckpt")

            run_epoch(session, model_val, test_data, tf.no_op(), vali=True)

        print("In validation, max accuracy is %lf, best RMSE is %lf, the epoch is %d, applying in test set." %
              (max_accuracy, best_rmse, record_epoch))

        print("Test:")
        model_val.input_hidden_layer_saver.restore(session, "./model/" + path + "/input_hidden_layer.ckpt")
        model_val.sent_lstm_saver.restore(session, "./model/" + path + "/sent_lstm.ckpt")
        model_val.sent_attention_saver.restore(session, "./model/" + path + "/sent_attention.ckpt")
        model_val.doc_lstm_saver.restore(session, "./model/" + path + "/doc_lstm.ckpt")
        model_val.doc_attention_saver.restore(session, "./model/" + path + "/doc_attention.ckpt")
        model_val.output_hidden_layer_saver.restore(session, "./model/" + path + "/output_hidden_layer.ckpt")
        acc_test, rmse_test = run_epoch(session, model_val, test_data, tf.no_op(), vali=True)
        print("Test accurary: %lf, Test RMSE: %lf" % (acc_test, rmse_test))


if __name__ == '__main__':

    path = "yelp13"
    val_data = json.load(open("./data/" + path + "/val_32.json", "r"))
    train_data = json.load(open("./data/" + path + "/train_32.json", "r"))
    test_data = json.load(open("./data/" + path + "/test_32.json", "r"))
    print("Finish loading data!")
    attention_data = json.load(open("./data/" + path + "/supervised_attention_32.json", "r"))
    voc_size = len(np.load("./data/" + path + "/word_embedding.npy"))
    config = Config("./data/" + path + "/word_embedding.npy", voc_size=voc_size)
    train(train_data, val_data, test_data, config, path, attention_data=attention_data)
